from obfsc8.src.obfsc8.get_columns_to_be_obfuscated \
    import get_columns_to_be_obfuscated
from test_data.test_json import test_json
from test_data.test_json_targeting_student_id \
    import test_json_targeting_student_id


def test_that_list_returned():
    restricted_fields = []
    result = get_columns_to_be_obfuscated(test_json, restricted_fields)

    assert isinstance(result, list)


def test_that_list_contains_correct_column_names():
    restricted_fields = []
    result = get_columns_to_be_obfuscated(test_json, restricted_fields)

    assert result == ["name", "email_address"]


def test_that_restricted_fields_correctly_removed_from_column_names():
    restricted_fields = ["student_id"]
    result1 = (get_columns_to_be_obfuscated
               (test_json_targeting_student_id, restricted_fields))

    assert result1 == ["name", "email_address"]

    restricted_fields = []
    result2 = (get_columns_to_be_obfuscated
               (test_json_targeting_student_id, restricted_fields))

    assert result2 == ["student_id", "name", "email_address"]



from obfsc8.src.obfsc8.get_s3_bucket_and_key_names \
    import get_s3_bucket_and_key_names
from test_data.test_json import test_json


def test_that_2_strings_returned():
    s3_bucket, s3_key = get_s3_bucket_and_key_names(test_json)

    assert isinstance(s3_bucket, str)
    assert isinstance(s3_key, str)


def test_that_s3_key_contains_period_but_s3_bucket_does_not():
    s3_bucket, s3_key = get_s3_bucket_and_key_names(test_json)

    assert '.' not in s3_bucket
    assert '.' in s3_key


def test_that_s3_key_and_bucket_names_match_those_in_json():
    s3_bucket, s3_key = get_s3_bucket_and_key_names(test_json)

    assert s3_bucket == "my_ingestion_bucket"
    assert s3_key == "new_data/file1.csv"



import boto3


def get_file_object_from_s3_bucket(bucket, key):
    """
    Retrieves file streaming object from S3 bucket

    Args:
        bucket: target S3 bucket
        key:    target S3 file key
    Returns:
        Requested file as streaming object
    """
    try:
        s3 = boto3.client("s3", region_name="eu-west-2")
        get_s3_file_object = s3.get_object(Bucket=bucket, Key=key)["Body"]

        return get_s3_file_object

    except Exception as e:
        print(f"File object retrieval from S3 bucket failed: {e}")



from .get_s3_bucket_and_key_names \
    import get_s3_bucket_and_key_names
from .get_file_object_from_s3_bucket \
    import get_file_object_from_s3_bucket
from .get_filetype import get_filetype
from .get_columns_to_be_obfuscated \
    import get_columns_to_be_obfuscated
from .obfuscate_csv_file import obfuscate_csv_file
from .obfuscate_parquet_file import obfuscate_parquet_file


def obfuscate(input_json, restricted_fields=[], replacement_string="***"):
    """
    Replaces all values within specified column/s, in CSV or Parquet file
    loaded from S3 bucket, with single replacement string, and writes
    resulting file in the same file format as a streamable object

    Args:
        input_json: JSON string detailing "file_to_obfuscate" and "pii_fields"
        , e.g.,
        '{
        "file_to_obfuscate": "s3://my_ingestion_bucket/new_data/file1.csv",
        "pii_fields": ["name", "email_address"]
        }'

        restricted_fields: list of fields that cannot be overwritten,
                            regardless of contents of input_json (default = [])

        replacement_string: string to be used to replace all values in the
                            specified PII fields (default = "***")
    Returns:
        BytesIO object containing obfuscated file data
    """

    s3_bucket_name, s3_key_name = get_s3_bucket_and_key_names(input_json)
    retrieved_file_object = (get_file_object_from_s3_bucket
                             (s3_bucket_name, s3_key_name))
    filetype = get_filetype(s3_key_name)
    columns_to_be_obfuscated = (get_columns_to_be_obfuscated
                                (input_json, restricted_fields))

    if filetype == "csv":
        transformed_file_data = (obfuscate_csv_file(retrieved_file_object,
                                                    columns_to_be_obfuscated,
                                                    replacement_string))

        return transformed_file_data

    elif filetype == 'parquet':
        transformed_file_data = (
            obfuscate_parquet_file(
                retrieved_file_object,
                columns_to_be_obfuscated,
                replacement_string))

        return transformed_file_data



import json


def get_s3_bucket_and_key_names(input_json):
    """
    Extracts S3 bucket and key names from "file_to_obfuscate"
    field in input JSON string

    Args:
        input_json: target JSON string

    Returns:
        s3_bucket:  string matching name of S3 bucket
        s3_key:     string matching name of S3 key
    """
    try:
        input_json_as_object = json.loads(input_json)

        input_filename = input_json_as_object["file_to_obfuscate"]
        input_filename_elements = input_filename.split('/')

        s3_bucket = input_filename_elements[2]
        s3_key = ('/').join(input_filename_elements[3:])

        return s3_bucket, s3_key

    except Exception as e:
        print(f'Could not read S3 bucket and key name from JSON: {e}')



def get_filetype(filename):
    """
    Extracts filetype from filename extension

    Args:
        filename: target filename
    Returns:
        String of filetype
    Raises:
        TypeError if filename is not a string
        ValueError if filename does not contain a period
        ValueError if determined filetype is not CSV or Parquet
    """
    if not isinstance(filename, str):
        raise TypeError('Input filename must be a string')
    if '.' not in filename:
        raise ValueError('Input filename must contain a period')

    try:
        filename_split_on_period = filename.split('.')
        filetype = filename_split_on_period[-1]

    except Exception as e:
        print(f'Failed to extract filetype from filename: {e}')

    if filetype not in ['csv', 'parquet']:
        raise ValueError(f'''Filetype {filetype} not recognised.  \
                   Filetype must be CSV or Parquet.''')

    return filetype



import pytest

from obfsc8.src.obfsc8.get_filetype import get_filetype


def test_that_string_returned():
    filename = "new_data/file1.csv"
    result = get_filetype(filename)

    assert isinstance(result, str)


def test_that_csv_returned_with_csv_file_input():
    csv_filename = "new_data/file1.csv"
    result = get_filetype(csv_filename)

    assert result == "csv"


def test_that_parquet_returned_with_parquet_file_input():
    parquet_filename = "new_data/file1.parquet"
    result = get_filetype(parquet_filename)

    assert result == "parquet"


def test_that_type_error_raised_if_input_filename_not_a_string():
    with pytest.raises(TypeError, match="must be a string"):
        get_filetype(674)


def test_that_value_error_raised_if_period_not_present_in_filename():
    with pytest.raises(ValueError, match="must contain a period"):
        get_filetype("new_data/file1")


def test_that_value_error_raised_if_filetype_not_CSV_or_Parquet():
    with pytest.raises(ValueError, match="Filetype must be CSV or Parquet"):
        get_filetype("new_data/file1.jpg")



import polars as pl
import io


def obfuscate_parquet_file(parquet_file_object,
                           columns_for_obfuscation,
                           replacement_string):
    """
    For a single Parquet file object input, replaces all values in
    specified columns with a single replacement string value,
    and returns bytes object of transformed Parquet data

    Args:
        parquet_file_object: target Parquet file object
        columns_for_obfuscation: list of target columns
        replacement_string: string to be used to replace
                              target column values
    Returns:
        BytesIO data object
    """
    try:
        parquet_obfuscation_operation = (
            pl.read_parquet(parquet_file_object).lazy()
            .with_columns(pl.col(columns_for_obfuscation)
                          .str.replace_all(r"(?s).*", replacement_string))
        )
        obfuscated_parquet_df = parquet_obfuscation_operation.collect()
        buffer = io.BytesIO()
        obfuscated_parquet_df.write_parquet(buffer)
        buffer.seek(0)

        return buffer

    except Exception as e:
        print(f'Failed to generate BytesIO object from Parquet file: {e}')



import polars.testing as pt
import polars as pl

from test_data.test_dataframe import test_dataframe


def test_that_s3_csv_written_and_retrieved_without_data_change(csv_from_s3):

    csv_file_from_s3_as_df = pl.read_csv(csv_from_s3)

    pt.assert_frame_equal(test_dataframe, csv_file_from_s3_as_df)



import polars as pl
import polars.testing as pt

from obfsc8.src.obfsc8.obfuscate_csv_file import obfuscate_csv_file
from test_data.test_dataframe import test_dataframe


columns_for_obfuscation = ["name", "email_address"]


def test_that_csv_file_returned_is_not_equivalent_to_the_file_input(
        csv_from_s3):

    buffer = obfuscate_csv_file(
        csv_from_s3, columns_for_obfuscation, "***")
    obfuscated_dataframe = pl.read_csv(buffer)

    pt.assert_frame_not_equal(test_dataframe, obfuscated_dataframe)


def test_that_all_values_in_non_target_columns_remain_unchanged(csv_from_s3):

    buffer = obfuscate_csv_file(
        csv_from_s3, columns_for_obfuscation, "***")
    obfuscated_dataframe = pl.read_csv(buffer)

    for column_name in obfuscated_dataframe.columns:
        if column_name not in columns_for_obfuscation:
            original_column_values = test_dataframe.get_column(column_name)
            obfuscated_column_values = (obfuscated_dataframe
                                        .get_column(column_name))

            (pt.assert_series_equal(original_column_values,
                                    obfuscated_column_values))


def test_that_all_values_in_target_columns_made_equal_to_replacement_string(
        csv_from_s3):

    buffer = obfuscate_csv_file(
        csv_from_s3, columns_for_obfuscation, "***")
    obfuscated_dataframe = pl.read_csv(buffer)

    obfuscated_column_values_list = []
    for column_name in columns_for_obfuscation:
        obfuscated_column_values = (obfuscated_dataframe
                                    .get_column(column_name))
        obfuscated_column_values_list.append(obfuscated_column_values)
        print(obfuscated_column_values_list)
    for i in range(1, len(obfuscated_column_values_list)):
        (pt.assert_series_equal(obfuscated_column_values_list[0],
                                obfuscated_column_values_list[i],
                                check_names=False))
    for j in range(0, len(obfuscated_column_values_list[0])):
        assert obfuscated_column_values_list[0][j] == "***"


def test_that_custom_string_leaves_non_target_columns_unchanged(csv_from_s3):

    buffer = obfuscate_csv_file(
        csv_from_s3, columns_for_obfuscation, "??")
    obfuscated_dataframe = pl.read_csv(buffer)

    for column_name in obfuscated_dataframe.columns:
        if column_name not in columns_for_obfuscation:
            original_column_values = test_dataframe.get_column(column_name)
            obfuscated_column_values = (obfuscated_dataframe
                                        .get_column(column_name))

            (pt.assert_series_equal(original_column_values,
                                    obfuscated_column_values))


def test_that_custom_replacement_string_replaces_values_in_target_columns(
        csv_from_s3):

    buffer = obfuscate_csv_file(
        csv_from_s3, columns_for_obfuscation, "??")
    obfuscated_dataframe = pl.read_csv(buffer)

    obfuscated_column_values_list = []
    for column_name in columns_for_obfuscation:
        obfuscated_column_values = (obfuscated_dataframe
                                    .get_column(column_name))
        obfuscated_column_values_list.append(obfuscated_column_values)
        print(obfuscated_column_values_list)
    for i in range(1, len(obfuscated_column_values_list)):
        (pt.assert_series_equal(obfuscated_column_values_list[0],
                                obfuscated_column_values_list[i],
                                check_names=False))
    for j in range(0, len(obfuscated_column_values_list[0])):
        assert obfuscated_column_values_list[0][j] == "??"
